import logging

from datetime import datetime
from psi_elt.abstract_elt_job import AbstractELTJob, LOADER_ELT_JOB_TYPE
from typeguard import typechecked
from typing import Any


@typechecked
class CustomLoader(AbstractELTJob):
    '''A CustomLoader is the second of the ELT jobs that is responsible for
    loading raw data into S3 in a queryable format.

    This class should be treated as abstract and should be extended.
    Child classes must override the _load function.
    '''

    def __init__(
        self,
        name: str,
        capture_time: datetime,
        aws_wrangler: Any = None,
        aws_session: Any = None,
        logger: logging.Logger = None
    ):
        super().__init__(
            name=name,
            elt_job_type=LOADER_ELT_JOB_TYPE,
            capture_time=capture_time,
            aws_session=aws_session,
            logger=logger
        )
        self.aws_wrangler = aws_wrangler

    def load(
        self,
        is_test: bool,
        load_kwargs: dict
    ):
        load_kwargs = self.remove_none_from_dict(load_kwargs)
        self._load(
            is_test,
            **load_kwargs
        )

    def _load(
        self,
        is_test: bool,
        **load_kwargs
    ) -> Any:
        raise Exception('CustomLoader._load must be overridden by the child class')



import logging

from datetime import datetime
from psi_elt.abstract_elt_job import AbstractELTJob, TRANSFORMER_ELT_JOB_TYPE
from typeguard import typechecked
from typing import Any, List, Union


@typechecked
class CustomTransformer(AbstractELTJob):
    '''A CustomTransformer is the third of the ELT jobs that is responsible for
    reading data (from S3, Athena, or Timestream) and transforming the data for
    use in a data mart.

    This class should be treated as abstract and should be extended.
    Child classes must override the _read_data and _transform functions.
    This class includes utilties for reading from S3, Athena, and Timestream.
    '''

    def __init__(
        self,
        name: str,
        capture_time: datetime,
        timestream_loader: Any = None,
        aws_session: Any = None,
        logger: logging.Logger = None,
        athena_api: Any = None,
        timestream_reader: Any = None
    ):
        super().__init__(
            name=name,
            elt_job_type=TRANSFORMER_ELT_JOB_TYPE,
            capture_time=capture_time,
            aws_session=aws_session,
            logger=logger
        )
        self.timestream_loader = timestream_loader
        self.athena_api = athena_api
        self.timestream_reader = timestream_reader

    def transform(
        self,
        is_test: bool,
        transform_kwargs: dict = None,
        timestream_load_kwargs: dict = None
    ):
        transform_kwargs = self.remove_none_from_dict(transform_kwargs)
        timestream_load_kwargs = self.remove_none_from_dict(timestream_load_kwargs)
        self._transform(
            is_test,
            timestream_load_kwargs=timestream_load_kwargs,
            **transform_kwargs
        )

    def _transform(
        self,
        is_test: bool,
        timestream_load_kwargs: dict = None,
        **transform_kwargs
    ):
        raise Exception('CustomTransformer._transform must be overridden by the child class')

    def write_to_timestream(
        self,
        data: list,
        capture_time: datetime,
        database_name: str = None,
        table_name: str = None,
        source: str = None,
        upsert: bool = False,
        ignore_parse_error: bool = False,
        max_batch_size: int = None,
        min_common_dimensions: int = None,
        is_multi: bool = True,
        is_preview: bool = False
    ):
        '''Calls TimestreamLoader.load() or load_preview() and raises an
        exception if there are any invalid items or rejected records
        '''
        timestream_load_input = {
            'data': data,
            'capture_time': capture_time,
            'database_name': database_name,
            'table_name': table_name,
            'source': source,
            'upsert': upsert,
            'ignore_parse_error': ignore_parse_error,
            'max_batch_size': max_batch_size,
            'min_common_dimensions': min_common_dimensions,
            'is_multi': is_multi
        }
        timestream_load_input = self.remove_none_from_dict(timestream_load_input)
        if is_preview:
            result = self.timestream_loader.load_preview(**timestream_load_input)
        else:
            result = self.timestream_loader.load(**timestream_load_input)
        
        if len(result['invalid_items']) > 0:
            raise Exception(f"Failed to write to Timestream with invalid items {result['invalid_items']}")
        if len(result['rejected_records']) > 0:
            raise Exception(f"Failed to write to Timestream with rejected records {result['rejected_records']}")

        



import logging
import pytz

from datetime import datetime, timedelta
from psi_elt import get_logger, TIME_UNITS
from psi_elt.abstract_elt_job import AbstractELTJob
from typeguard import typechecked
from typing import Any, Dict

NoneType = type(None)


class UnrecognizedELTJobException(Exception):
    pass


@typechecked
class AbstractRunner:
    def __init__(
        self,
        jobs: Dict[str, Any],
        logger: logging.Logger = None
    ):
        '''
        Initializes an AbstractRunner.

        This class should be treated as abstract and should be extended.
        Child classes must override the _run function. Optionally, child
        classes may override the parse_event and validate_parsed_event
        functions as well.

        This class includes utilties for getting a capture time and
        initializing an ELT job.
        '''
        assert jobs, 'A jobs dict is required to initialize an AbstractRunner'
        self.__jobs = jobs
        self.__logger = logger or get_logger()

    @property
    def logger(self) -> logging.Logger:
        return self.__logger

    def run(self, event: Dict[str, Any]) -> Dict[str, Any]:
        t0 = datetime.now()
        parsed_event = self.parse_event(event)
        self.validate_parsed_event(parsed_event)
        self.logger.info(f'Running with parsed event: {parsed_event}')
        job = self._run(**parsed_event)
        self.logger.info(f'ELT Job completed in {str(datetime.now() - t0)}')
        return {
            'event': event,
            'parsed_event': parsed_event,
            'job': job.to_dict()
        }

    def _run(self, job: AbstractELTJob, **parsed_event):
        raise Exception('AbstractRunner._run should be overridden by the child class')

    def parse_event(self, event: Dict[str, Any]) -> Dict[str, Any]:
        raise Exception('AbstractRunner.parse_event should be overridden by the child class')

    def validate_parsed_event(self, parsed_event: Dict[str, Any]):
        raise Exception('AbstractRunner.validate_parsed_event should be overridden by the child class')

    def _validate_dict(self, dct: Dict[str, Any], schema: Dict[str, Any]):
        required_keys = [
            k for k, v in schema.items()
            if not((isinstance(v, list) and NoneType in v) or isinstance(v, dict))
        ]
        optional_keys = [
            k for k, v in schema.items()
            if (isinstance(v, list) and NoneType in v) or isinstance(v, dict)
        ]

        missing_keys = [k for k in required_keys if k not in dct]
        assert not missing_keys, f'Missing keys: {missing_keys}'

        extra_keys = [k for k in dct if k not in required_keys + optional_keys]
        assert not extra_keys, f'Extra keys: {extra_keys}'

        for k, v in dct.items():
            required_type = schema[k]
            if isinstance(required_type, list):
                assert type(v) in required_type, f'Invalid {k}, expected {required_type} but got: {v}'
            elif isinstance(required_type, dict):
                self._validate_dict(v, required_type)
            else:
                assert type(v) == required_type, f'Invalid {k}, expected {required_type} but got: {v}'

    def get_capture_time(self, event: dict) -> datetime:
        input_options: dict = event.get('input_options') or {}
        if 'capture_time' in event and event['capture_time']:
            # fromisoformat returns a naive datetime object that IS NOT adjusted for the local timezone
            try:
                capture_time = datetime.fromisoformat(
                    event['capture_time']
                ).replace(tzinfo=pytz.utc)
            except Exception:
                capture_time = self._parse_ago_time(event['capture_time'])
        elif input_options.get('s3_uri'):
            capture_time_ms = event['s3_uri'].split('/')[-1]
            # fromtimestamp returns a naive datetime object that IS adjusted for the local timezone
            capture_time = datetime.fromtimestamp(
                int(capture_time_ms)/1000
            ).astimezone().astimezone(pytz.utc)
        else:
            # now returns a naive datetime object that IS adjusted for the local timezone
            capture_time = datetime.now().astimezone().astimezone(pytz.utc)

        assert isinstance(capture_time, datetime), f'Invalid capture time. Expected a datetime object, but got: {capture_time}'
        assert capture_time.tzinfo == pytz.UTC, f'Invalid capture time. Expected capture time to be UTC, but got: {capture_time}'
        return capture_time

    def _parse_ago_time(self, capture_time: str):
        try:
            if capture_time[-2:].isalpha():
                time_interval = TIME_UNITS[capture_time[-2:]]
                number = capture_time[0:-2]
            elif capture_time[-1:].isalpha():
                time_interval = TIME_UNITS[capture_time[-1:]]
                number = capture_time[0:-1]
            else:
                raise Exception(f'Unable to parse {capture_time}')
            timedelta_params = {time_interval: int(number)}
            return datetime.now().astimezone().astimezone(pytz.UTC)-timedelta(
                **timedelta_params
            )
        except Exception as e:
            self.__logger.error(f'Failed to parse ago time "{capture_time}" with exception: {e}')
            raise e

    def get_job(
        self,
        job_name: str,
        capture_time: datetime,
        job_init_kwargs: dict
    ) -> AbstractELTJob:
        cls = self.__jobs.get(job_name)
        if cls:
            return cls(
                job_name,
                capture_time,
                **job_init_kwargs
            )
        raise UnrecognizedELTJobException(f'Unrecognized ELT job with name: {job_name}')



import logging

from datetime import datetime
from psi_elt.abstract_runner import AbstractRunner, NoneType
from psi_elt.custom_transformer import CustomTransformer
from typeguard import typechecked
from typing import Any, Dict


@typechecked
class TransformerRunner(AbstractRunner):
    def __init__(
        self,
        transformers: Dict[str, Any],
        timestream_loader: Any = None,
        logger: logging.Logger = None
    ):
        '''
        Initializes a TransformerRunner.

        Accepts a transformers dict, mapping transformer name to transformer class.
        For example:
        {
            "tag-compliance": TagComplianceTransformer,
            "qualys": QualysTransformer
        }
        '''
        super().__init__(jobs=transformers, logger=logger)
        self.__timestream_loader = timestream_loader

    def parse_event(self, event: Dict[str, Any]) -> Dict[str, Any]:
        '''Parses a transformer event. Child classes may override this function to
        implement custom event parsing rules.

        #### Required event keys
           1. name : the name of the transformer to run

        #### Optional event keys
          1. capture_time                         : target time for the transformer to run against
          2. is_test                              : boolean for whethor or not this is a test run
          3. input_options.athena_workgroup       : Athena workgroup to use when reading data
          4. input_options.athena_data_catalog    : Athena data catalog to use when reading data
          5. input_options.athena_database        : Athena database to read data from
          6. input_options.athena_table           : Athena table to read data from
          7. output_options.database_name         : Timestream database to write data to
          8. output_options.table_name            : Timestream table to write data to
          9. output_options.write_to_timestream   : boolean for whether to write to Timestream
         10. output_options.upsert                : boolean for whether to upsert or insert data
         11. output_options.max_batch_size        : int max batch size used when batching
         12. output_options.min_common_dimensions : int min common dimensions used when batching
         13. output_options.is_multi              : boolean for whether to use multi-measure records
        '''
        parsed_event = {
            'name': event.get('name'),
            'capture_time': self.get_capture_time(event),
            'job_init_kwargs': {},
            'is_test': bool(event.get('is_test')),
            'transform_kwargs': {},
            'timestream_load_kwargs': {}
        }

        # Parse input options
        if isinstance(event.get('input_options'), dict):
            input_options: dict = event['input_options']
            parsed_event['transform_kwargs'] = {
                k: v for k, v in input_options.items()
                if v is not None and k in [
                    'athena_workgroup',
                    'athena_data_catalog',
                    'athena_database',
                    'athena_table'
                ]
            }

        # Parse output options
        if isinstance(event.get('output_options'), dict):
            output_options: dict = event['output_options']
            parsed_event['timestream_load_kwargs'] = {
                k: v for k, v in output_options.items()
                if v is not None and k in [
                    'database_name',
                    'table_name',
                    'upsert',
                    'max_batch_size',
                    'min_common_dimensions',
                    'is_multi'
                ]
            }
            if 'write_to_timestream' in output_options:
                parsed_event['transform_kwargs']['write_to_timestream'] = output_options['write_to_timestream']
        return parsed_event

    def validate_parsed_event(self, parsed_event: Dict[str, Any]):
        '''Validates that a parsed event has the required format'''
        schema = {
            'name': str,
            'capture_time': datetime,
            'job_init_kwargs': {},
            'is_test': bool,
            'transform_kwargs': {
                'athena_workgroup': [str, NoneType],
                'athena_data_catalog': [str, NoneType],
                'athena_database': [str, NoneType],
                'athena_table': [str, NoneType],
                'write_to_timestream': [bool, NoneType]
            },
            'timestream_load_kwargs': {
                'database_name': [str, NoneType],
                'table_name': [str, NoneType],
                'upsert': [bool, NoneType],
                'ignore_parse_error': [bool, NoneType],
                'max_batch_size': [int, NoneType],
                'min_common_dimensions': [int, NoneType],
                'is_multi': [bool, NoneType]
            }
        }
        try:
            self._validate_dict(parsed_event, schema)
        except AssertionError as e:
            self.logger.error('Invalid event')
            raise e

    def _run(
        self,
        name: str,
        capture_time: datetime,
        job_init_kwargs: dict,
        is_test: bool,
        transform_kwargs: dict,
        timestream_load_kwargs: dict
    ) -> CustomTransformer:
        job_init_kwargs['timestream_loader'] = self.__timestream_loader
        transformer: CustomTransformer = self.get_job(name, capture_time, job_init_kwargs)
        transformer.transform(
            is_test,
            transform_kwargs,
            timestream_load_kwargs
        )
        return transformer



import logging

from datetime import datetime
from psi_elt.abstract_runner import AbstractRunner, NoneType
from psi_elt.custom_extractor import CustomExtractor
from typeguard import typechecked
from typing import Any, Dict


@typechecked
class ExtractorRunner(AbstractRunner):
    def __init__(
        self,
        extractors: Dict[str, Any],
        s3_loader: Any = None,
        logger: logging.Logger = None
    ):
        '''
        Initializes a ExtractorRunner.

        Accepts a extractors dict, mapping extractor name to extractor class.
        For example:
        {
            "divvy": DivvyExtractor,
            "pagerduty": PagerDutyExtractor
        }
        '''
        super().__init__(jobs=extractors, logger=logger)
        self.__s3_loader = s3_loader

    def parse_event(self, event: Dict[str, Any]) -> Dict[str, Any]:
        '''Parses an extractor event. Child classes may override this function to
        implement custom event parsing rules.

        #### Required event keys
           1. name : the name of the extractor to run

        #### Optional event keys
           1. capture_time               : target time for the job to run against
           2. is_test                    : boolean for whethor or not this is a test run
           3. input_options.account_id   : AWS account ID to extract data from
           4. output_options.s3_uri      : S3 location to write data to
           5. output_options.s3_path     : S3 path to write data to
           6. output_options.write_to_s3 : boolean for whether or not to write to S3
        '''
        parsed_event = {
            'name': event.get('name'),
            'capture_time': self.get_capture_time(event),
            'job_init_kwargs': {},
            'is_test': bool(event.get('is_test')),
            'extract_kwargs': {},
            's3_load_kwargs': {}
        }

        # Parse input options
        if isinstance(event.get('input_options'), dict):
            input_options: dict = event['input_options']
            parsed_event['extract_kwargs'] = {
                k: v for k, v in input_options.items()
                if v is not None and k in [
                    'account_id'
                ]
            }

        # Parse output options
        if isinstance(event.get('output_options'), dict):
            output_options: dict = event['output_options']
            parsed_event['s3_load_kwargs'] = {
                k: v for k, v in output_options.items()
                if v is not None and k in [
                    's3_uri',
                    's3_path'
                ]
            }
            if 'write_to_s3' in output_options:
                parsed_event['extract_kwargs']['write_to_s3'] = output_options['write_to_s3']
        return parsed_event

    def validate_parsed_event(self, parsed_event: Dict[str, Any]):
        '''Validates that a parsed event has the required format'''
        schema = {
            'name': str,
            'capture_time': datetime,
            'job_init_kwargs': {},
            'is_test': bool,
            'extract_kwargs': {
                'account_id': [str, NoneType],
                'write_to_s3': [bool, NoneType]
            },
            's3_load_kwargs': {
                's3_uri': [str, NoneType],
                's3_path': [str, NoneType]
            }
        }
        try:
            self._validate_dict(parsed_event, schema)
        except AssertionError as e:
            self.logger.error('Invalid event')
            raise e

    def _run(
        self,
        name: str,
        capture_time: datetime,
        job_init_kwargs: dict,
        is_test: bool,
        extract_kwargs: dict,
        s3_load_kwargs: dict
    ) -> CustomExtractor:
        job_init_kwargs['s3_loader'] = self.__s3_loader
        extractor: CustomExtractor = self.get_job(name, capture_time, job_init_kwargs)
        extractor.extract(
            is_test,
            extract_kwargs,
            s3_load_kwargs
        )
        return extractor
